# cython: cpow=True

import re
import warnings
import numpy as np
from scipy._lib._util import check_random_state, _transition_to_rng

cimport numpy as np
cimport cython
from cython.view cimport array
from libc.math cimport sqrt, sin, cos, atan2, acos, hypot, isnan, NAN, pi

np.import_array()

# utilities for empty array initialization
cdef inline double[:] _empty1(int n) noexcept:
    if n == 0:
        return array(shape=(1,), itemsize=sizeof(double), format=b"d")[:0]
    return array(shape=(n,), itemsize=sizeof(double), format=b"d")
cdef inline double[:, :] _empty2(int n1, int n2) noexcept :
    if n1 == 0:
        return array(shape=(1, n2), itemsize=sizeof(double), format=b"d")[:0]
    return array(shape=(n1, n2), itemsize=sizeof(double), format=b"d")
cdef inline double[:, :, :] _empty3(int n1, int n2, int n3) noexcept:
    if n1 == 0:
        return array(shape=(1, n2, n3), itemsize=sizeof(double), format=b"d")[:0]
    return array(shape=(n1, n2, n3), itemsize=sizeof(double), format=b"d")
cdef inline double[:, :] _zeros2(int n1, int n2) noexcept:
    if n1 == 0:
        return array(shape=(1, n2), itemsize=sizeof(double), format=b"d")[:0]
    cdef double[:, :] arr = array(shape=(n1, n2),
        itemsize=sizeof(double), format=b"d")
    arr[:, :] = 0
    return arr

# flat implementations of numpy functions
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline double[:] _cross3(const double[:] a, const double[:] b) noexcept:
    cdef double[:] result = _empty1(3)
    result[0] = a[1]*b[2] - a[2]*b[1]
    result[1] = a[2]*b[0] - a[0]*b[2]
    result[2] = a[0]*b[1] - a[1]*b[0]
    return result

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline double _dot3(const double[:] a, const double[:] b) noexcept nogil:
    return a[0]*b[0] + a[1]*b[1] + a[2]*b[2]

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline double _norm3(const double[:] elems) noexcept nogil:
    return sqrt(_dot3(elems, elems))

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline double _normalize4(double[:] elems) noexcept nogil:
    cdef double norm = sqrt(_dot3(elems, elems) + elems[3]*elems[3])

    if norm == 0:
        return NAN

    elems[0] /= norm
    elems[1] /= norm
    elems[2] /= norm
    elems[3] /= norm

    return norm

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int _argmax4(const double[:] a) noexcept nogil:
    cdef int imax = 0
    cdef double vmax = a[0]

    for i in range(1, 4):
        if a[i] > vmax:
            imax = i
            vmax = a[i]

    return imax

ctypedef unsigned char uchar

cdef double[3] _ex = [1, 0, 0]
cdef double[3] _ey = [0, 1, 0]
cdef double[3] _ez = [0, 0, 1]

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline const double[:] _elementary_basis_vector(uchar axis) noexcept:
    if axis == b'x': return _ex
    elif axis == b'y': return _ey
    elif axis == b'z': return _ez

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int _elementary_basis_index(uchar axis) noexcept:
    if axis == b'x': return 0
    elif axis == b'y': return 1
    elif axis == b'z': return 2

# Reduce the quaternion double coverage of the rotation group to a unique
# canonical "positive" single cover
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void _quat_canonical_single(double[:] q) noexcept nogil:
    if ((q[3] < 0)
        or (q[3] == 0 and q[0] < 0)
        or (q[3] == 0 and q[0] == 0 and q[1] < 0)
        or (q[3] == 0 and q[0] == 0 and q[1] == 0 and q[2] < 0)):
        q[0] *= -1.0
        q[1] *= -1.0
        q[2] *= -1.0
        q[3] *= -1.0

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void _quat_canonical(double[:, :] q) noexcept:
    cdef Py_ssize_t n = q.shape[0]
    for ind in range(n):
        _quat_canonical_single(q[ind])

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int _get_angles(
    double[:] angles, bint extrinsic, bint symmetric, bint sign,
    double lamb, double a, double b, double c, double d) noexcept:
    # Returns 1 if a gimbal warning is detected and 0 otherwise (so that
    # warnings can be handled at a higher level)

    # intrinsic/extrinsic conversion helpers
    cdef int angle_first, angle_third
    if extrinsic:
        angle_first = 0
        angle_third = 2
    else:
        angle_first = 2
        angle_third = 0

    cdef double half_sum, half_diff
    cdef int case

    # Step 2
    # Compute second angle...
    angles[1] = 2 * atan2(hypot(c, d), hypot(a, b))

    # ... and check if equal to is 0 or pi, causing a singularity
    if abs(angles[1]) <= 1e-7:
        case = 1
    elif abs(angles[1] - <double>pi) <= 1e-7:
        case = 2
    else:
        case = 0 # normal case

    # Step 3
    # compute first and third angles, according to case
    half_sum = atan2(b, a)
    half_diff = atan2(d, c)

    if case == 0:  # no singularities
        angles[angle_first] = half_sum - half_diff
        angles[angle_third] = half_sum + half_diff

    else:  # any degenerate case
        angles[2] = 0
        if case == 1:
            angles[0] = 2 * half_sum
        else:
            angles[0] = 2 * half_diff * (-1 if extrinsic else 1)

    # for Tait-Bryan/asymmetric sequences
    if not symmetric:
        angles[angle_third] *= sign
        angles[1] -= lamb

    for idx in range(3):
        if angles[idx] < -pi:
            angles[idx] += 2 * pi
        elif angles[idx] > pi:
            angles[idx] -= 2 * pi

    if case != 0:
        return 1
    return 0


@cython.boundscheck(False)
@cython.wraparound(False)
cdef double[:, :] _compute_euler_from_quat(
    np.ndarray[double, ndim=2] quat,
    const uchar[:] seq,
    bint extrinsic,
    bint suppress_warnings
) noexcept:
    # The algorithm assumes extrinsic frame transformations. The algorithm
    # in the paper is formulated for rotation quaternions, which are stored
    # directly by Rotation.
    # Adapt the algorithm for our case by reversing both axis sequence and
    # angles for intrinsic rotations when needed

    if not extrinsic:
        seq = seq[::-1]

    cdef int i = _elementary_basis_index(seq[0])
    cdef int j = _elementary_basis_index(seq[1])
    cdef int k = _elementary_basis_index(seq[2])

    cdef bint symmetric = i == k
    if symmetric:
        k = 3 - i - j # get third axis

    # Step 0
    # Check if permutation is even (+1) or odd (-1)
    cdef int sign = (i - j) * (j - k) * (k - i) // 2

    cdef Py_ssize_t num_rotations = quat.shape[0]

    # some forward definitions
    cdef double a, b, c, d
    cdef double[:, :] angles = _empty2(num_rotations, 3)
    cdef int n_gimbal_warnings = 0

    for ind in range(num_rotations):

        # Step 1
        # Permutate quaternion elements
        if symmetric:
            a = quat[ind, 3]
            b = quat[ind, i]
            c = quat[ind, j]
            d = quat[ind, k] * sign
        else:
            a = quat[ind, 3] - quat[ind, j]
            b = quat[ind, i] + quat[ind, k] * sign
            c = quat[ind, j] + quat[ind, 3]
            d = quat[ind, k] * sign - quat[ind, i]

        n_gimbal_warnings += _get_angles(
            angles[ind], extrinsic, symmetric, sign, pi / 2, a, b, c, d
        )

    if n_gimbal_warnings > 0 and not suppress_warnings:
        warnings.warn("Gimbal lock detected. Setting third angle to zero "
                      "since it is not possible to uniquely determine "
                      "all angles.", stacklevel=2)

    return angles

@cython.boundscheck(False)
@cython.wraparound(False)
cdef double[:, :] _compute_davenport_from_quat(
    np.ndarray[double, ndim=2] quat, np.ndarray[double, ndim=1] n1,
    np.ndarray[double, ndim=1] n2, np.ndarray[double, ndim=1] n3,
    bint extrinsic,
    bint suppress_warnings
):
    # The algorithm assumes extrinsic frame transformations. The algorithm
    # in the paper is formulated for rotation quaternions, which are stored
    # directly by Rotation.
    # Adapt the algorithm for our case by reversing both axis sequence and
    # angles for intrinsic rotations when needed

    if not extrinsic:
        n1, n3 = n3, n1

    cdef double[:] n_cross = _cross3(n1, n2)
    cdef double lamb = atan2(_dot3(n3, n_cross), _dot3(n3, n1))

    cdef int correct_set = False
    if lamb < 0:
        # alternative set of angles compatible with as_euler implementation
        n2 = -n2
        lamb = -lamb
        n_cross[0] = -n_cross[0]
        n_cross[1] = -n_cross[1]
        n_cross[2] = -n_cross[2]
        correct_set = True

    cdef double[:] quat_lamb = np.array([
            sin(lamb / 2) * n2[0],
            sin(lamb / 2) * n2[1],
            sin(lamb / 2) * n2[2],
            cos(lamb / 2)]
    )

    cdef Py_ssize_t num_rotations = quat.shape[0]

    # some forward definitions
    cdef double[:, :] angles = _empty2(num_rotations, 3)
    cdef double[:] quat_transformed = _empty1(4)
    cdef double a, b, c, d
    cdef int n_gimbal_warnings = 0

    for ind in range(num_rotations):
        _compose_quat_single(quat_lamb, quat[ind], quat_transformed)

        # Step 1
        # Permutate quaternion elements
        a = quat_transformed[3]
        b = _dot3(quat_transformed[:3], n1)
        c = _dot3(quat_transformed[:3], n2)
        d = _dot3(quat_transformed[:3], n_cross)

        n_gimbal_warnings += _get_angles(angles[ind], extrinsic, False, 1, lamb, a, b, c, d)

        if correct_set:
            angles[ind, 1] = -angles[ind, 1]

    if n_gimbal_warnings > 0 and not suppress_warnings:
        warnings.warn("Gimbal lock detected. Setting third angle to zero "
                      "since it is not possible to uniquely determine "
                      "all angles.", stacklevel=2)

    return angles

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void _compose_quat_single( # calculate p * q into r
    const double[:] p, const double[:] q, double[:] r
) noexcept:
    cdef double[:] cross = _cross3(p[:3], q[:3])

    r[0] = p[3]*q[0] + q[3]*p[0] + cross[0]
    r[1] = p[3]*q[1] + q[3]*p[1] + cross[1]
    r[2] = p[3]*q[2] + q[3]*p[2] + cross[2]
    r[3] = p[3]*q[3] - p[0]*q[0] - p[1]*q[1] - p[2]*q[2]

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline double[:, :] _compose_quat(
    const double[:, :] p, const double[:, :] q
) noexcept:
    cdef Py_ssize_t n = q.shape[0] if p.shape[0] == 1 else p.shape[0]
         
    cdef double[:, :] product = _empty2(n, 4)

    # dealing with broadcasting
    if p.shape[0] == 1:
        for ind in range(n):
            _compose_quat_single(p[0], q[ind], product[ind])
    elif q.shape[0] == 1:
        for ind in range(n):
            _compose_quat_single(p[ind], q[0], product[ind])
    else:
        for ind in range(n):
            _compose_quat_single(p[ind], q[ind], product[ind])

    return product

def compose_quat(p, q):
    """Composition of quaternions."""
    return _compose_quat(p, q)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline double[:, :] _make_elementary_quat(
    uchar axis, const double[:] angles
) noexcept:
    cdef Py_ssize_t n = angles.shape[0]
    cdef double[:, :] quat = _zeros2(n, 4)

    cdef int axis_ind
    if axis == b'x':   axis_ind = 0
    elif axis == b'y': axis_ind = 1
    elif axis == b'z': axis_ind = 2

    for ind in range(n):
        quat[ind, 3] = cos(angles[ind] / 2)
        quat[ind, axis_ind] = sin(angles[ind] / 2)
    return quat

@cython.boundscheck(False)
@cython.wraparound(False)
cdef double[:, :] _elementary_quat_compose(
    const uchar[:] seq, const double[:, :] angles, bint intrinsic=False
) noexcept:
    cdef double[:, :] result = _make_elementary_quat(seq[0], angles[:, 0])
    cdef Py_ssize_t seq_len = seq.shape[0]

    for idx in range(1, seq_len):
        if intrinsic:
            result = _compose_quat(
                result,
                _make_elementary_quat(seq[idx], angles[:, idx]))
        else:
            result = _compose_quat(
                _make_elementary_quat(seq[idx], angles[:, idx]),
                result)
    return result


@cython.boundscheck(False)
@cython.wraparound(False)
def from_quat(double[:, :] quat, bint normalize=True, bint copy=True, bint scalar_first=False):
    if quat.ndim != 2 or quat.shape[1] != 4:
        raise ValueError(f"Expected `quat` to have shape (N, 4), got {quat.shape}.")

    cdef Py_ssize_t num_rotations = quat.shape[0]

    if num_rotations > 0:  # Avoid 0-sized axis errors
        if scalar_first:
            quat = np.roll(quat, -1, axis=1)
        elif normalize or copy:
            quat = quat.copy()

        if normalize:
            for ind in range(num_rotations):
                if isnan(_normalize4(quat[ind, :])):
                    raise ValueError("Found zero norm quaternions in `quat`.")

    return np.asarray(quat, dtype=float)


@cython.boundscheck(False)
@cython.wraparound(False)
def from_euler(seq, angles, bint degrees=False):
    num_axes = len(seq)
    if num_axes < 1 or num_axes > 3:
        raise ValueError("Expected axis specification to be a non-empty "
                            "string of upto 3 characters, got {}".format(seq))

    intrinsic = (re.match(r'^[XYZ]{1,3}$', seq) is not None)
    extrinsic = (re.match(r'^[xyz]{1,3}$', seq) is not None)
    if not (intrinsic or extrinsic):
        raise ValueError("Expected axes from `seq` to be from ['x', 'y', "
                            "'z'] or ['X', 'Y', 'Z'], got {}".format(seq))

    if any(seq[i] == seq[i+1] for i in range(num_axes - 1)):
        raise ValueError("Expected consecutive axes to be different, "
                            "got {}".format(seq))

    seq = seq.lower()
    
    angles = np.asarray(angles, dtype=float)

    if angles.ndim > 2:  # The backend should never be called with these inputs
        raise ValueError("Cython backend is only compatible with up to 2D inputs")

    if degrees:
        angles = np.deg2rad(angles)

    is_single = angles.ndim < 2
    angles = np.atleast_2d(angles)  # Ensure 0, 1 and 2D arrays are 2D
    
    if angles.shape[1] != num_axes:
        raise ValueError("Expected last dimension of `angles` to match number of"
                         f" sequence axes specified, got {angles.shape[1]}.")

    quat = _elementary_quat_compose(seq.encode(), angles, intrinsic)

    if is_single:
        return np.asarray(quat, dtype=float)[0]
    return np.asarray(quat, dtype=float)


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def from_matrix(matrix, bint assume_valid=False):
    cdef int ind
    is_single = False
    matrix = np.array(matrix, dtype=float)

    if (matrix.ndim not in [2, 3] or
        matrix.shape[len(matrix.shape)-2:] != (3, 3)):
        raise ValueError("Expected `matrix` to have shape (3, 3) or "
                            "(N, 3, 3), got {}".format(matrix.shape))

    # If a single matrix is given, convert it to 3D 1 x 3 x 3 matrix but
    # set is_single to True so that we can return appropriate objects in
    # the `to_...` methods
    if matrix.shape == (3, 3):
        matrix = matrix[np.newaxis, :, :]
        is_single = True

    if not assume_valid:
        # Calculate the determinant of the rotation matrix
        # (should be positive for right-handed rotations)
        dets = np.linalg.det(matrix)
        if np.any(dets <= 0):
            ind = np.where(dets <= 0)[0][0]
            raise ValueError("Non-positive determinant (left-handed or null "
                                f"coordinate frame) in rotation matrix {ind}: "
                                f"{matrix[ind]}.")

        # Gramian orthogonality check
        # (should be the identity matrix for orthogonal matrices)
        # Note that we have already ruled out left-handed cases above
        gramians = matrix @ np.transpose(matrix, (0, 2, 1))
        is_orthogonal = np.all(np.isclose(gramians, np.eye(3), atol=1e-12),
                                axis=(1, 2))
        indices_to_orthogonalize = np.where(~is_orthogonal)[0]

        # Orthogonalize the rotation matrices where necessary
        if len(indices_to_orthogonalize) > 0:
            # Exact solution to the orthogonal Procrustes problem using singular
            # value decomposition
            U, _, Vt = np.linalg.svd(matrix[indices_to_orthogonalize, :, :])
            matrix[indices_to_orthogonalize, :, :] = U @ Vt

    # Convert the orthogonal rotation matrices to quaternions using the
    # algorithm described in [3]_. This will also apply another
    # orthogonalization step to correct for any small errors in the matrices
    # that skipped the SVD step above.
    if is_single:
        return _from_matrix_orthogonal(matrix[0])
    return _from_matrix_orthogonal(matrix)


@cython.boundscheck(False)
@cython.wraparound(False)
def _from_matrix_orthogonal(matrix):
    is_single = False
    if matrix.shape == (3, 3):
        matrix = matrix[np.newaxis, :, :]
        is_single = True

    cdef double[:, :, :] cmatrix
    cmatrix = matrix
    cdef Py_ssize_t num_rotations = cmatrix.shape[0]
    cdef Py_ssize_t i, j, k
    cdef double[:] decision = _empty1(4)
    cdef int choice
    cdef int ind

    cdef double[:, :] quat = _empty2(num_rotations, 4)

    for ind in range(num_rotations):
        decision[0] = cmatrix[ind, 0, 0]
        decision[1] = cmatrix[ind, 1, 1]
        decision[2] = cmatrix[ind, 2, 2]
        decision[3] = cmatrix[ind, 0, 0] + cmatrix[ind, 1, 1] \
                    + cmatrix[ind, 2, 2]
        choice = _argmax4(decision)

        if choice != 3:
            i = choice
            j = (i + 1) % 3
            k = (j + 1) % 3

            quat[ind, i] = 1 - decision[3] + 2 * cmatrix[ind, i, i]
            quat[ind, j] = cmatrix[ind, j, i] + cmatrix[ind, i, j]
            quat[ind, k] = cmatrix[ind, k, i] + cmatrix[ind, i, k]
            quat[ind, 3] = cmatrix[ind, k, j] - cmatrix[ind, j, k]
        else:
            quat[ind, 0] = cmatrix[ind, 2, 1] - cmatrix[ind, 1, 2]
            quat[ind, 1] = cmatrix[ind, 0, 2] - cmatrix[ind, 2, 0]
            quat[ind, 2] = cmatrix[ind, 1, 0] - cmatrix[ind, 0, 1]
            quat[ind, 3] = 1 + decision[3]

        # normalize
        _normalize4(quat[ind])

    if is_single:
        return np.asarray(quat, dtype=float)[0]
    return np.asarray(quat, dtype=float)


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def from_rotvec(rotvec, bint degrees=False):
    is_single = False
    rotvec = np.asarray(rotvec, dtype=float)
    if degrees:
        rotvec = np.deg2rad(rotvec)

    if rotvec.ndim not in [1, 2] or rotvec.shape[len(rotvec.shape)-1] != 3:
        raise ValueError("Expected `rot_vec` to have shape (3,) "
                            "or (N, 3), got {}".format(rotvec.shape))

    # If a single vector is given, convert it to a 2D 1 x 3 matrix but
    # set self._single to True so that we can return appropriate objects
    # in the `as_...` methods
    cdef double[:, :] crotvec
    if rotvec.shape == (3,):
        crotvec = rotvec[None, :]
        is_single = True
    else:
        crotvec = rotvec

    cdef Py_ssize_t num_rotations = crotvec.shape[0]
    cdef double angle, scale, angle2
    cdef double[:, :] quat = _empty2(num_rotations, 4)

    for ind in range(num_rotations):
        angle = _norm3(crotvec[ind, :])

        if angle <= 1e-3:  # small angle Taylor series expansion
            angle2 = angle * angle
            scale = 0.5 - angle2 / 48 + angle2 * angle2 / 3840
        else:  # large angle
            scale = sin(angle / 2) / angle

        quat[ind, 0] = scale * crotvec[ind, 0]
        quat[ind, 1] = scale * crotvec[ind, 1]
        quat[ind, 2] = scale * crotvec[ind, 2]
        quat[ind, 3] = cos(angle / 2)

    if is_single:
        return np.asarray(quat, dtype=float)[0]
    return np.asarray(quat, dtype=float)


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def from_mrp(mrp):
    is_single = False
    mrp = np.asarray(mrp, dtype=float)

    if mrp.ndim not in [1, 2] or mrp.shape[len(mrp.shape) - 1] != 3:
        raise ValueError("Expected `mrp` to have shape (3,) "
                            "or (N, 3), got {}".format(mrp.shape))

    # If a single vector is given, convert it to a 2D 1 x 3 matrix but
    # set self._single to True so that we can return appropriate objects
    # in the `as_...` methods
    cdef double[:, :] cmrp
    if mrp.shape == (3,):
        cmrp = mrp[None, :]
        is_single = True
    else:
        cmrp = mrp

    cdef Py_ssize_t num_rotations = cmrp.shape[0]
    cdef double[:, :] quat = _empty2(num_rotations, 4)
    cdef double mrp_squared_plus_1

    for ind in range(num_rotations):
        mrp_squared_plus_1 = 1 + _dot3(cmrp[ind, :], cmrp[ind, :])

        quat[ind, 0] = 2 * cmrp[ind, 0] / mrp_squared_plus_1
        quat[ind, 1] = 2 * cmrp[ind, 1] / mrp_squared_plus_1
        quat[ind, 2] = 2 * cmrp[ind, 2] / mrp_squared_plus_1
        quat[ind, 3] = (2 - mrp_squared_plus_1) / mrp_squared_plus_1

    if is_single:
        return quat[0]
    return quat
    

@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def as_quat(double[:, :] quat, bint canonical=False, *, bint scalar_first=False) -> double[:, :]:
    q = np.array(quat, copy=True)
    if canonical:
        _quat_canonical(q)

    if scalar_first:
        q = np.roll(q, 1, axis=1)

    return q


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def as_matrix(double[:, :] quat) -> double[:, :, :]:

    cdef Py_ssize_t num_rotations = quat.shape[0]
    cdef double[:, :, :] matrix = _empty3(num_rotations, 3, 3)

    cdef double x, y, z, w, x2, y2, z2, w2
    cdef double xy, zw, xz, yw, yz, xw

    for ind in range(num_rotations):
        x = quat[ind, 0]
        y = quat[ind, 1]
        z = quat[ind, 2]
        w = quat[ind, 3]

        x2 = x * x
        y2 = y * y
        z2 = z * z
        w2 = w * w

        xy = x * y
        zw = z * w
        xz = x * z
        yw = y * w
        yz = y * z
        xw = x * w

        matrix[ind, 0, 0] = x2 - y2 - z2 + w2
        matrix[ind, 1, 0] = 2 * (xy + zw)
        matrix[ind, 2, 0] = 2 * (xz - yw)

        matrix[ind, 0, 1] = 2 * (xy - zw)
        matrix[ind, 1, 1] = - x2 + y2 - z2 + w2
        matrix[ind, 2, 1] = 2 * (yz + xw)

        matrix[ind, 0, 2] = 2 * (xz + yw)
        matrix[ind, 1, 2] = 2 * (yz - xw)
        matrix[ind, 2, 2] = - x2 - y2 + z2 + w2

    return np.asarray(matrix)


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def as_rotvec(double[:, :] quat, bint degrees=False) -> double[:, :]:
    cdef Py_ssize_t num_rotations = len(quat)
    cdef double angle, scale, angle2
    cdef double[:, :] rotvec = _empty2(num_rotations, 3)
    cdef double[:] quat_single

    for ind in range(num_rotations):
        quat_single = quat[ind, :].copy()
        _quat_canonical_single(quat_single)  # w > 0 ensures that 0 <= angle <= pi

        angle = 2 * atan2(_norm3(quat_single), quat_single[3])

        if angle <= 1e-3:  # small angle Taylor series expansion
            angle2 = angle * angle
            scale = 2 + angle2 / 12 + 7 * angle2 * angle2 / 2880
        else:  # large angle
            scale = angle / sin(angle / 2)

        rotvec[ind, 0] = scale * quat_single[0]
        rotvec[ind, 1] = scale * quat_single[1]
        rotvec[ind, 2] = scale * quat_single[2]

    if degrees:
        rotvec = np.rad2deg(rotvec)

    return np.asarray(rotvec)


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def as_mrp(double[:, :] quat) -> double[:, :]:
    cdef Py_ssize_t num_rotations = len(quat)
    cdef double[:, :] mrps = _empty2(num_rotations, 3)
    cdef int sign
    cdef double denominator

    # TODO: Check if broadcasting is possible similar to the xp backend
    for ind in range(num_rotations):

        # Ensure we are calculating the set of MRPs that correspond
        # to a rotation of <= 180
        sign = -1 if quat[ind, 3] < 0 else 1

        denominator = 1 + sign * quat[ind, 3]
        for i in range(3):
            mrps[ind, i] = sign * quat[ind, i] / denominator

    return np.asarray(mrps)


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def as_euler(
    double[:, :] quat,
    seq,
    bint degrees=False,
    bint suppress_warnings=False
) -> double[:, :]:
    # Prepare axis sequence to call Euler angles conversion algorithm.
    if len(seq) != 3:
        raise ValueError("Expected 3 axes, got {}.".format(seq))

    intrinsic = (re.match(r'^[XYZ]{1,3}$', seq) is not None)
    extrinsic = (re.match(r'^[xyz]{1,3}$', seq) is not None)
    if not (intrinsic or extrinsic):
        raise ValueError("Expected axes from `seq` to be from "
                            "['x', 'y', 'z'] or ['X', 'Y', 'Z'], "
                            "got {}".format(seq))

    if any(seq[i] == seq[i+1] for i in range(2)):
        raise ValueError("Expected consecutive axes to be different, "
                            "got {}".format(seq))

    seq = seq.lower()
    if not extrinsic:
        seq = seq[::-1]

    # The algorithm assumes extrinsic frame transformations. The algorithm
    # in the paper is formulated for rotation quaternions, which are stored
    # directly by Rotation. See:
    # Bernardes E, Viollet S (2022) Quaternion to Euler angles conversion: A
    # direct, general and computationally efficient method.
    # PLoS ONE 17(11): e0276302. https://doi.org/10.1371/journal.pone.0276302
    cdef const uchar[:] cseq = seq.encode()
    cdef int i = _elementary_basis_index(cseq[0])
    cdef int j = _elementary_basis_index(cseq[1])
    cdef int k = _elementary_basis_index(cseq[2])

    cdef bint symmetric = i == k
    if symmetric:
        k = 3 - i - j # get third axis

    # Step 0
    # Check if permutation is even (+1) or odd (-1)
    cdef int sign = (i - j) * (j - k) * (k - i) // 2

    cdef Py_ssize_t num_rotations = quat.shape[0]

    # some forward definitions
    cdef double a, b, c, d
    cdef double[:, :] angles = _empty2(num_rotations, 3)
    cdef int n_gimbal_warnings = 0

    for ind in range(num_rotations):

        # Step 1
        # Permutate quaternion elements
        if symmetric:
            a = quat[ind, 3]
            b = quat[ind, i]
            c = quat[ind, j]
            d = quat[ind, k] * sign
        else:
            a = quat[ind, 3] - quat[ind, j]
            b = quat[ind, i] + quat[ind, k] * sign
            c = quat[ind, j] + quat[ind, 3]
            d = quat[ind, k] * sign - quat[ind, i]

        n_gimbal_warnings += _get_angles(
            angles[ind], extrinsic, symmetric, sign, pi / 2, a, b, c, d
        )

    if n_gimbal_warnings > 0 and not suppress_warnings:
        warnings.warn("Gimbal lock detected. Setting third angle to zero "
                      "since it is not possible to uniquely determine "
                      "all angles.", stacklevel=2)

    if degrees:
        angles = np.rad2deg(angles)
    return np.asarray(angles)


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def as_davenport(
    double[:, :] quat,
    double[:, :] axes,
    order,
    bint degrees=False,
    bint suppress_warnings=False
) -> double[:, :]:
    cdef np.ndarray[double, ndim=2] q = np.asarray(quat)

    cdef bint extrinsic
    if order in ['e', 'extrinsic']:
        extrinsic = True
    elif order in ['i', 'intrinsic']:
        extrinsic = False
    else:
        raise ValueError("order should be 'e'/'extrinsic' for extrinsic "
                            "sequences or 'i'/'intrinsic' for intrinsic "
                            "sequences, got {}".format(order))

    if len(axes) != 3:
        raise ValueError("Expected 3 axes, got {}.".format(len(axes)))

    if axes.ndim != 2 or axes.shape[1] != 3:
        raise ValueError("Axes must be vectors of length 3.")

    cdef np.ndarray[double, ndim=1] n1, n2, n3
    n1, n2, n3 = np.asarray(axes)

    # normalize axes
    n1 = n1 / np.linalg.norm(n1)
    n2 = n2 / np.linalg.norm(n2)
    n3 = n3 / np.linalg.norm(n3)

    if np.dot(n1, n2) >= 1e-7:
        raise ValueError("Consecutive axes must be orthogonal.")
    if np.dot(n2, n3) >= 1e-7:
        raise ValueError("Consecutive axes must be orthogonal.")

    angles = np.asarray(_compute_davenport_from_quat(
        q, n1, n2, n3, extrinsic, suppress_warnings
    ))

    if degrees:
        angles = np.rad2deg(angles)

    return angles


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def inv(double[:, :] quat) -> double[:, :]:
    cdef np.ndarray[double, ndim=2] q_inv = np.array(quat, copy=True)
    q_inv[:, 0] *= -1
    q_inv[:, 1] *= -1
    q_inv[:, 2] *= -1
    return q_inv


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@_transition_to_rng('random_state', position_num=1)
def random(num: int | None = None, rng=None, shape: int | tuple[int, ...] | None = None):
    rng = check_random_state(rng)
    if num is None and shape is None:
        shape = ()
    elif num is not None:
        shape = (num,)
    elif isinstance(shape, int):
        shape = (shape,)
    elif not isinstance(shape, tuple):
        raise ValueError("`shape` must be an int or a tuple of ints or None.")
    return rng.normal(size=shape + (4,))


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def identity(num: int | None, shape: int | tuple[int, ...] | None = None):
    if num is None and shape is None:
        shape = ()
    elif num is not None:
        shape = (num,)
    elif isinstance(shape, int):
        shape = (shape,)
    elif not isinstance(shape, tuple):
        raise ValueError("`shape` must be an int or a tuple of ints or None.")
    q = np.zeros(shape + (4,), dtype=np.float64)
    q[..., 3] = 1
    return q


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def magnitude(double[:, :] quat) -> double[:, :]:
    cdef Py_ssize_t num_rotations = quat.shape[0]
    cdef double[:] angles = _empty1(num_rotations)

    for ind in range(num_rotations):
        angles[ind] = 2 * atan2(_norm3(quat[ind, :3]), abs(quat[ind, 3]))

    return np.asarray(angles)


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def approx_equal(double[:, :] quat, double[:, :] other, atol = None, bint degrees=False) -> bool[:]:
    len_self = len(quat)
    len_other = len(other)
    if not(len_self == 1 or len_other == 1 or len_self == len_other):
        raise ValueError(
            f"Expected broadcastable shapes in both rotations, got {len_self} "
            f"rotations in first and {len_other} rotations in second object."
        )

    if atol is None:
        if degrees:
            warnings.warn("atol must be set to use the degrees flag, "
                            "defaulting to 1e-8 radians.")
        atol = 1e-8  # radians
    elif degrees:
        atol = np.deg2rad(atol)

    angles = magnitude(compose_quat(other, inv(quat)))
    return angles < atol


@cython.embedsignature(True)
@cython.boundscheck(False)
def mean(double[:, :] quat, weights=None, axis=None):
    if quat.shape[0] == 0:
        raise ValueError("Mean of an empty rotation set is undefined.")
    # The Cython path assumes quat is Nx4, so axis has to be None, 0, -1, (0,), (-1,),
    # or (). The code path is unchanged for any of the options except (), where we
    # immediately return the quaternion
    if axis == ():
        return quat

    if axis is None:
        axis = (0,)
    if isinstance(axis, int):
        axis = (axis,)
    if not isinstance(axis, tuple):  # Must be tuple by now
        raise ValueError("`axis` must be None, int, or tuple of ints.")
    if min(axis) < -1 or max(axis) > 0:
        raise ValueError(
            f"axis {axis} is out of bounds for rotation with shape "
            f"{np.asarray(quat).shape[:-1]}."
        )
    # Axis must be 0 for the cython backend. Everything else should have raised an
    # error during validation.
    axis = (0,)

    if weights is None:
        weights = np.ones(quat.shape[0])
    else:
        weights = np.asarray(weights)
        if np.any(weights < 0):
            raise ValueError("`weights` must be non-negative.")
        if weights.ndim != 1:
            raise ValueError(f"Expected `weights` to be 1 dimensional, got "
                             f"{weights.shape}.")
        if weights.shape[0] != quat.shape[0]:
            raise ValueError("Expected `weights` to have number of values equal to "
                             f"number of rotations, got {weights.shape[0]} values and "
                             f"{quat.shape[0]} rotations.")

    quat = np.asarray(quat)
    K = np.dot(weights * quat.T, quat)
    _, v = np.linalg.eigh(K)
    return v[:, -1]


@cython.embedsignature(True)
@cython.boundscheck(False)
def reduce(double[:, :] quat, left=None, right=None):
    if left is None and right is None:
        reduced = quat
        return reduced, None, None
    elif right is None:
        right = identity(1)
    elif left is None:
        left = identity(1)

    # Levi-Civita tensor for triple product computations
    e = np.zeros((3, 3, 3))
    e[0, 1, 2] = e[1, 2, 0] = e[2, 0, 1] = 1
    e[0, 2, 1] = e[2, 1, 0] = e[1, 0, 2] = -1

    # We want to calculate the real components of q = l * p * r. It can
    # be shown that:
    #     qs = ls * ps * rs - ls * dot(pv, rv) - ps * dot(lv, rv)
    #          - rs * dot(lv, pv) - dot(cross(lv, pv), rv)
    # where ls and lv denote the scalar and vector components of l.

    def split_rotation(q):
        q = np.atleast_2d(q)
        return q[:, -1], q[:, :-1]

    p = quat
    ps, pv = split_rotation(p)
    ls, lv = split_rotation(left)
    rs, rv = split_rotation(right)

    qs = np.abs(np.einsum('i,j,k', ls, ps, rs) -
                np.einsum('i,jx,kx', ls, pv, rv) -
                np.einsum('ix,j,kx', lv, ps, rv) -
                np.einsum('ix,jx,k', lv, pv, rs) -
                np.einsum('xyz,ix,jy,kz', e, lv, pv, rv))
    qs = np.reshape(np.moveaxis(qs, 1, 0), (qs.shape[1], -1))

    # Find best indices from scalar components
    max_ind = np.argmax(np.reshape(qs, (len(qs), -1)), axis=1)
    left_best = max_ind // len(rv)
    right_best = max_ind % len(rv)

    # Reduce the rotation using the best indices
    left = left[left_best]
    right = right[right_best]
    reduced = _compose_quat(left, _compose_quat(p, right))

    return reduced, left_best, right_best


@cython.embedsignature(True)
@cython.boundscheck(False)
def apply(double[:, :] quat, double[:, :] vectors, bint inverse=False) -> double[:, :]:
    cdef Py_ssize_t n_vectors = len(vectors)
    cdef Py_ssize_t n_rotations = len(quat)

    if n_vectors != 1 and n_rotations != 1 and n_vectors != n_rotations:
        raise ValueError(
            f"Cannot broadcast {n_rotations} rotations to {n_vectors} vectors."
        )

    cdef np.ndarray matrix = as_matrix(quat)

    if inverse:
        matrix = np.swapaxes(matrix, -1, -2)

    if n_vectors == 1 and n_rotations == 1:
        return np.matmul(matrix, vectors[0])

    if n_rotations == 1:
        # Single rotation/many vectors, use matmul for speed: The axes argument
        # is such that the input arguments don't need to be transposed and the
        # output argument is continuous in memory.
        return np.matmul(matrix, vectors, axes=[(-2, -1), (-1, -2), (-1, -2)])[0]

    # for stacks of matrices einsum is faster
    return np.einsum('ijk,ik->ij', matrix, vectors)


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def setitem(quat: double[:, :], value: double[:, :], indexer):
    quat[indexer] = value
    return quat


@cython.embedsignature(True)
def align_vectors(a, b, weights=None, bint return_sensitivity=False):
    # Check input vectors
    a_original = np.array(a, dtype=float)
    b_original = np.array(b, dtype=float)
    a = np.atleast_2d(a_original)
    b = np.atleast_2d(b_original)
    if a.shape[-1] != 3:
        raise ValueError("Expected input `a` to have shape (3,) or "
                            "(N, 3), got {}".format(a_original.shape))
    if b.shape[-1] != 3:
        raise ValueError("Expected input `b` to have shape (3,) or "
                            "(N, 3), got {}".format(b_original.shape))
    if a.shape != b.shape:
        raise ValueError("Expected inputs `a` and `b` to have same shapes"
                            ", got {} and {} respectively.".format(
                                a_original.shape, b_original.shape))
    N = len(a)

    # Check weights
    if weights is None:
        weights = np.ones(N)
    else:
        weights = np.array(weights, dtype=float)
        if weights.ndim != 1:
            raise ValueError("Expected `weights` to be 1 dimensional, got "
                                "shape {}.".format(weights.shape))
        if N > 1 and (weights.shape[0] != N):
            raise ValueError("Expected `weights` to have number of values "
                                "equal to number of input vectors, got "
                                "{} values and {} vectors.".format(
                                weights.shape[0], N))
        if (weights < 0).any():
            raise ValueError("`weights` may not contain negative values")

    # For the special case of a single vector pair, we use the infinite
    # weight code path
    if N == 1:
        weight_is_inf = np.array([True])
    else:
        weight_is_inf = np.isposinf(weights)
    n_inf = np.sum(weight_is_inf)

    # Check for an infinite weight, which indicates that the corresponding
    # vector pair is the primary unmoving reference to which we align the
    # other secondary vectors
    if n_inf > 1:
        raise ValueError("Only one infinite weight is allowed")
    elif n_inf == 1:  # TODO: Refactor into _align_vectors_fixed as in the xp backend
        if return_sensitivity:
            raise ValueError("Cannot return sensitivity matrix with an "
                                "infinite weight or one vector pair")
        a_pri = a[weight_is_inf]
        b_pri = b[weight_is_inf]
        a_pri_norm = _norm3(a_pri[0])
        b_pri_norm = _norm3(b_pri[0])
        if a_pri_norm == 0 or b_pri_norm == 0:
            raise ValueError("Cannot align zero length primary vectors")
        a_pri /= a_pri_norm
        b_pri /= b_pri_norm

        # We first find the minimum angle rotation between the primary
        # vectors.
        cross = np.cross(b_pri[0], a_pri[0])
        cross_norm = _norm3(cross)
        theta = atan2(cross_norm, _dot3(a_pri[0], b_pri[0]))
        tolerance = 1e-3  # tolerance for small angle approximation (rad)
        q_flip = np.array([[0.0, 0.0, 0.0, 1.0]])
        if (np.pi - theta) < tolerance:
            # Near pi radians, the Taylor series approximation of x/sin(x)
            # diverges, so for numerical stability we flip pi and then
            # rotate back by the small angle pi - theta
            if cross_norm == 0:
                # For antiparallel vectors, cross = [0, 0, 0] so we need to
                # manually set an arbitrary orthogonal axis of rotation
                i = np.argmin(np.abs(a_pri[0]))
                r = np.zeros(3)
                r[i - 1], r[i - 2] = a_pri[0][i - 2], -a_pri[0][i - 1]
            else:
                r = cross  # Shortest angle orthogonal axis of rotation
            q_flip = from_rotvec((r / np.linalg.norm(r) * np.pi)[None, ...])
            theta = np.pi - theta
            cross = -cross
        if abs(theta) < tolerance:
            # Small angle Taylor series approximation for numerical stability
            theta2 = theta * theta
            r = cross * (1 + theta2 / 6 + theta2 * theta2 * 7 / 360)
        else:
            r = cross * theta / np.sin(theta)
        q_pri = _compose_quat(from_rotvec(r[None, :]), q_flip)

        if N == 1:
            # No secondary vectors, so we are done
            q_opt = q_pri
        else:
            a_sec = a[~weight_is_inf]
            b_sec = b[~weight_is_inf]
            weights_sec = weights[~weight_is_inf]

            # We apply the first rotation to the b vectors to align the
            # primary vectors, resulting in vectors c. The secondary
            # vectors must now be rotated about that primary vector to best
            # align c to a.
            c_sec = apply(q_pri, b_sec)

            # Calculate vector components for the angle calculation. The
            # angle phi to rotate a single 2D vector C to align to 2D
            # vector A in the xy plane can be found with the equation
            # phi = atan2(cross(C, A), dot(C, A))
            #     = atan2(|C|*|A|*sin(phi), |C|*|A|*cos(phi))
            # The below equations perform the same operation, but with the
            # 3D rotation restricted to the 2D plane normal to a_pri, where
            # the secondary vectors are projected into that plane. We then
            # find the composite angle from the weighted sum of the
            # axial components in that plane, minimizing the 2D alignment
            # problem.
            # Note that einsum('ij,ij->i', X, Y) is the row-wise dot
            # product of X and Y.
            sin_term = np.einsum('ij,ij->i', np.cross(c_sec, a_sec), a_pri)
            cos_term = (np.einsum('ij,ij->i', c_sec, a_sec)
                        - (np.einsum('ij,ij->i', c_sec, a_pri)
                            * np.einsum('ij,ij->i', a_sec, a_pri)))
            phi = atan2(np.sum(weights_sec * sin_term),
                        np.sum(weights_sec * cos_term))
            q_sec = from_rotvec((phi * a_pri[0])[None, :])

            # Compose these to get the optimal rotation
            q_opt = _compose_quat(q_sec, q_pri)

        # Calculated the root sum squared distance. We force the error to
        # be zero for the infinite weight vectors since they will align
        # exactly.
        weights_inf_zero = weights.copy()
        if N > 1 or np.isposinf(weights[0]):
            # Skip non-infinite weight single vectors pairs, we used the
            # infinite weight code path but don't want to zero that weight
            weights_inf_zero[weight_is_inf] = 0
        a_est = apply(q_opt, b)
        rssd = np.sqrt(np.sum(weights_inf_zero @ (a - a_est)**2))

        return q_opt[0, ...], rssd, None

    # If no infinite weights and multiple vectors, proceed with normal
    # algorithm
    # Note that einsum('ji,jk->ik', X, Y) is equivalent to np.dot(X.T, Y)
    B = np.einsum('ji,jk->ik', weights[:, None] * a, b)
    u, s, vh = np.linalg.svd(B)

    # Correct improper rotation if necessary (as in Kabsch algorithm)
    if np.linalg.det(u @ vh) < 0:
        s[-1] = -s[-1]
        u[:, -1] = -u[:, -1]

    C = np.dot(u, vh)

    if s[1] + s[2] < 1e-16 * s[0]:
        warnings.warn("Optimal rotation is not uniquely or poorly defined "
                        "for the given sets of vectors.")

    rssd = np.sqrt(max(
        np.sum(weights * np.sum(b ** 2 + a ** 2, axis=1)) - 2 * np.sum(s),
        0))

    if return_sensitivity:
        zeta = (s[0] + s[1]) * (s[1] + s[2]) * (s[2] + s[0])
        kappa = s[0] * s[1] + s[1] * s[2] + s[2] * s[0]
        with np.errstate(divide='ignore', invalid='ignore'):
            sensitivity = np.mean(weights) / zeta * (
                    kappa * np.eye(3) + np.dot(B, B.T))
        return _from_matrix_orthogonal(C), rssd, sensitivity
    return _from_matrix_orthogonal(C), rssd, None


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def pow(double[:, :] quat, n) -> double[:, :]:
    if isinstance(n, np.ndarray) and n.ndim != 0 and n.shape != (1,):
        raise ValueError("Array exponent must be a scalar")
    # Exact short-cuts
    if n == 0:
        return identity(quat.shape[0])
    elif n == -1:
        return inv(quat)
    elif n == 1:
        return quat
    # general scaling of rotation angle
    return from_rotvec(n * as_rotvec(quat))


@cython.embedsignature(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def from_davenport(axes, order: str, angles, bint degrees=False):
    if order in ['e', 'extrinsic']:
        extrinsic = True
    elif order in ['i', 'intrinsic']:
        extrinsic = False
    else:
        raise ValueError("order should be 'e'/'extrinsic' for extrinsic "
                            "sequences or 'i'/'intrinsic' for intrinsic "
                            "sequences, got {}".format(order))

    axes = np.asarray(axes)
    if axes.ndim == 1:
        axes = axes.reshape([1, 3])

    num_axes = len(axes)

    if axes.shape[1] != 3:
        raise ValueError("Axes must be vectors of length 3.")

    if num_axes < 1 or num_axes > 3:
        raise ValueError("Expected up to 3 axes, got {}".format(num_axes))

    # normalize axes
    norm = np.repeat(np.linalg.norm(axes, axis=1), 3)
    axes = axes / norm.reshape(num_axes, 3)

    if (num_axes > 1 and abs(np.dot(axes[0], axes[1])) >= 1e-7 or
        num_axes > 2 and abs(np.dot(axes[1], axes[2])) >= 1e-7):
        raise ValueError("Consecutive axes must be orthogonal.")

    angles = np.asarray(angles, dtype=float)
    if degrees:
        angles = np.deg2rad(angles)

    is_single = angles.ndim < 2
    if angles.ndim != 2:
        angles = np.atleast_2d(angles)

    if angles.shape[1] != axes.shape[0]:
        raise ValueError("Expected `angles` to match number of axes, got "
                         f"{angles.shape[1]} angles and {axes.shape[0]} axes.")

    q = identity(len(angles))
    for i in range(num_axes):
        qi = from_rotvec(angles[:, i, np.newaxis] * axes[i])
        if extrinsic:
            q = compose_quat(qi, q)
        else:
            q = compose_quat(q, qi)

    if is_single:
        return np.asarray(q, dtype=float)[0]
    return np.asarray(q, dtype=float)



"""
A collection of tests to make sure that wrapped namespaces agree with the bare ones
on whether to return a view or a copy of inputs.
"""
import pytest
from ._helpers import import_, wrapped_libraries


FUNC_INPUTS = [
    # func_name, arr_input, dtype,  scalar_value
    ('abs', [1, 2], 'int8', 3),
    ('abs', [1, 2], 'float32', 3.),
    ('ceil', [1, 2], 'int8', 3),
    ('clip', [1, 2], 'int8', 3),
    ('conj', [1, 2], 'int8', 3),
    ('floor', [1, 2], 'int8', 3),
    ('imag', [1j, 2j], 'complex64', 3),
    ('positive', [1, 2], 'int8', 3),
    ('real', [1., 2.], 'float32', 3.),
    ('round', [1, 2], 'int8', 3),
    ('sign', [0, 0], 'float32', 3),
    ('trunc', [1, 2], 'int8', 3),
    ('trunc', [1, 2], 'float32', 3),
]


def ensure_unary(func, arr):
    """Make a trivial unary function from func."""
    if func.__name__ == 'clip':
        return lambda x: func(x, arr[0], arr[1])
    return func


def is_view(func, a, value):
    """Apply `func`, mutate the output; does the input change?"""
    b = func(a)
    b[0] = value
    return a[0] == value


@pytest.mark.parametrize('xp_name', wrapped_libraries + ['array_api_strict'])
@pytest.mark.parametrize('inputs', FUNC_INPUTS, ids=[inp[0] for inp in FUNC_INPUTS])
def test_view_or_copy(inputs, xp_name):
    bare_xp = import_(xp_name, wrapper=False)
    wrapped_xp = import_(xp_name, wrapper=True)

    func_name, arr_input, dtype_str, value = inputs
    dtype = getattr(bare_xp, dtype_str)

    bare_func = getattr(bare_xp, func_name)
    bare_func = ensure_unary(bare_func, arr_input)

    wrapped_func = getattr(wrapped_xp, func_name)
    wrapped_func = ensure_unary(wrapped_func, arr_input)

    # bare namespace: mutate the output, does the input change?
    a = bare_xp.asarray(arr_input, dtype=dtype)
    is_view_bare = is_view(bare_func, a, value)

    # wrapped namespace: mutate the output, does the input change?
    a1 = wrapped_xp.asarray(arr_input, dtype=dtype)
    is_view_wrapped = is_view(wrapped_func, a1, value)

    assert is_view_bare == is_view_wrapped



import warnings

import numpy as np

from .common import Benchmark, XPBenchmark, is_xslow, safe_import

with safe_import():
    from scipy.cluster.hierarchy import linkage, is_isomorphic
    from scipy.cluster.vq import kmeans, kmeans2, vq, whiten


class Linkage(XPBenchmark):
    method = ['single', 'complete', 'average', 'weighted', 'centroid', 'median', 'ward']
    param_names = (*XPBenchmark.param_names, "size", "method")
    if is_xslow():
        size = [100, 180, 325, 585, 1054, 1898, 3420, 6162, 11101, 20000]
    else:
        size = [2000]
    params = (*XPBenchmark.params, size, method)

    def setup(self, backend, size, method):
        super().setup(backend, linkage, static_argnames="method")

        rng = np.random.default_rng(0)
        y = self.xp.asarray(rng.standard_normal((size, 2)))
        self.y = self.synchronize(y)

        if self.warmup:
            self.func(self.y, method=method)

    def time_linkage(self, backend, size, method):
        self.func(self.y, method=method)


class IsIsomorphic(XPBenchmark):
    NCLUSTERS = 5
    # This is very slow and memory intensive, but necessary to
    # let _most_ backends approach O(n*logn) behaviour.
    # Note: memory usage = 16 * nobs
    if is_xslow():
        nobs = [100, 1_000, 10_000, 100_000, 1_000_000, 10_000_000, 100_000_000]
    else:
        nobs = [100, 100_000]
    # Skip cpu backends for nobs greater than this. 
    # They should all have reached O(n*logn) behaviour by then.
    CPU_MAX_OBS = 1_000_000

    param_names = (*XPBenchmark.param_names, "nobs")
    params = (*XPBenchmark.params, nobs)

    def setup(self, backend, nobs):
        use_cuda = backend == "cupy" or backend.endswith(":cuda")
        if not use_cuda and nobs > self.CPU_MAX_OBS:
            raise NotImplementedError("Skipping huge size on CPU")

        super().setup(backend, is_isomorphic)

        rng = np.random.default_rng(0)
        a = self.xp.asarray(rng.integers(0, self.NCLUSTERS, size=nobs))
        p = self.xp.asarray(rng.permutation(self.NCLUSTERS))
        b = self.xp.take(p, a)
        self.a, self.b = self.synchronize(a, b)

        if self.warmup:
            self.func(self.a, self.b)

    def time_is_isomorphic(self, backend, nobs):
        self.func(self.a, self.b)


class KMeans(Benchmark):
    params = [2, 10, 50]
    param_names = ['k']

    def __init__(self):
        rnd = np.random.RandomState(0)
        self.obs = rnd.rand(1000, 5)

    def time_kmeans(self, k):
        kmeans(self.obs, k, iter=10)


class KMeans2(Benchmark):
    params = [[2, 10, 50], ['random', 'points', '++']]
    param_names = ['k', 'init']

    def __init__(self):
        rnd = np.random.RandomState(0)
        self.obs = rnd.rand(1000, 5)

    def time_kmeans2(self, k, init):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                ("One of the clusters is empty. Re-run kmeans with a "
                 "different initialization"),
                UserWarning)
            kmeans2(self.obs, k, minit=init, iter=10)


class VQ(Benchmark):
    params = [[2, 10, 50], ['float32', 'float64']]
    param_names = ['k', 'dtype']

    def __init__(self):
        rnd = np.random.RandomState(0)
        self.data = rnd.rand(5000, 5)
        self.cbook_source = rnd.rand(50, 5)

    def setup(self, k, dtype):
        self.obs = self.data.astype(dtype)
        self.cbook = self.cbook_source[:k].astype(dtype)

    def time_vq(self, k, dtype):
        vq(self.obs, self.cbook)


class Whiten(XPBenchmark):
    if is_xslow():
        shape = [(10, 10), (32, 32), (100, 100), (320, 320),
                 (1000, 1000), (3200, 3200), (10_000, 10_000)]
    else:
        shape = [(10, 10), (100, 100)]

    param_names = (*XPBenchmark.param_names, "shape")
    params = (*XPBenchmark.params, shape)

    def setup(self, backend, shape):
        super().setup(backend, whiten, static_argnames="check_finite")

        rng = np.random.default_rng(0)
        obs = self.xp.asarray(rng.uniform(0, 100.0, size=shape))
        self.obs = self.synchronize(obs)

        if self.warmup:
            self.func(self.obs, check_finite=False)

    def time_whiten(self, backend, shape):
        self.func(self.obs, check_finite=False)



import subprocess
import sys
import warnings

import numpy as np
import pytest

import array_api_compat
from array_api_compat import array_namespace

from ._helpers import all_libraries, wrapped_libraries, xfail


@pytest.mark.parametrize("use_compat", [True, False, None])
@pytest.mark.parametrize(
    "api_version", [None, "2021.12", "2022.12", "2023.12", "2024.12"]
)
@pytest.mark.parametrize("library", all_libraries)
def test_array_namespace(request, library, api_version, use_compat):
    xp = pytest.importorskip(library)

    array = xp.asarray([1.0, 2.0, 3.0])
    if use_compat and library not in wrapped_libraries:
        pytest.raises(ValueError, lambda: array_namespace(array, use_compat=use_compat))
        return
    if (library == "sparse" and api_version in ("2023.12", "2024.12")) or (
        library == "jax.numpy" and api_version in ("2021.12", "2022.12", "2023.12")
    ):
        xfail(request, "Unsupported API version")

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', UserWarning)
        namespace = array_namespace(array, api_version=api_version, use_compat=use_compat)

    if use_compat is False or use_compat is None and library not in wrapped_libraries:
        if library == "jax.numpy" and not hasattr(xp, "__array_api_version__"):
            # Backwards compatibility for JAX <0.4.32
            import jax.experimental.array_api
            assert namespace == jax.experimental.array_api
        else:
            assert namespace == xp
    elif library == "dask.array":
        assert namespace == array_api_compat.dask.array
    else:
        assert namespace == getattr(array_api_compat, library)

    if library == "numpy":
        # check that the same namespace is returned for NumPy scalars
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)

            scalar_namespace = array_namespace(
                xp.float64(0.0), api_version=api_version, use_compat=use_compat
            )
            assert scalar_namespace == namespace


def test_jax_backwards_compat():
    """On JAX <0.4.32, test that array_namespace works even if
    jax.experimental.array_api has not been imported yet.
    """
    pytest.importorskip("jax")
    code = """\
import sys
import jax.numpy
import array_api_compat

array = jax.numpy.asarray([1.0, 2.0, 3.0])
assert 'jax.experimental.array_api' not in sys.modules
namespace = array_api_compat.array_namespace(array)

if hasattr(jax.numpy, '__array_api_version__'):
    assert namespace == jax.numpy
else:
    import jax.experimental.array_api
    assert namespace == jax.experimental.array_api
"""
    subprocess.check_call([sys.executable, "-c", code])


def test_jax_zero_gradient():
    jax = pytest.importorskip("jax")
    jx = jax.numpy.arange(4)
    jax_zero = jax.vmap(jax.grad(jax.numpy.float32, allow_int=True))(jx)
    assert array_namespace(jax_zero) is array_namespace(jx)


def test_array_namespace_errors():
    pytest.raises(TypeError, lambda: array_namespace([1]))
    pytest.raises(TypeError, lambda: array_namespace())

    x = np.asarray([1, 2])
    pytest.raises(TypeError, lambda: array_namespace((x, x)))
    pytest.raises(TypeError, lambda: array_namespace(x, (x, x)))


@pytest.mark.parametrize("library", all_libraries)
def test_array_namespace_many_args(library):
    xp = pytest.importorskip(library)
    a = xp.asarray(1)
    b = xp.asarray(2)
    assert array_namespace(a, b) is array_namespace(a)


def test_array_namespace_mismatch():
    xp = pytest.importorskip("array_api_strict")
    with pytest.raises(TypeError, match="Multiple namespaces"):
        array_namespace(np.asarray(1), xp.asarray(1))


def test_get_namespace():
    # Backwards compatible wrapper
    assert array_api_compat.get_namespace is array_namespace


@pytest.mark.parametrize("library", all_libraries)
def test_python_scalars(library):
    xp = pytest.importorskip(library)
    a = xp.asarray([1, 2])
    xp = array_namespace(a)

    pytest.raises(TypeError, lambda: array_namespace(1))
    pytest.raises(TypeError, lambda: array_namespace(1.0))
    pytest.raises(TypeError, lambda: array_namespace(1j))
    pytest.raises(TypeError, lambda: array_namespace(True))
    pytest.raises(TypeError, lambda: array_namespace(None))

    assert array_namespace(a, 1) is xp
    assert array_namespace(a, 1.0) is xp
    assert array_namespace(a, 1j) is xp
    assert array_namespace(a, True) is xp
    assert array_namespace(a, None) is xp



"""
Convenience interface to N-D interpolation

.. versionadded:: 0.9

"""
import numpy as np
from ._interpnd import (LinearNDInterpolator, NDInterpolatorBase,
     CloughTocher2DInterpolator, _ndim_coords_from_arrays)
from scipy.spatial import cKDTree

__all__ = ['griddata', 'NearestNDInterpolator', 'LinearNDInterpolator',
           'CloughTocher2DInterpolator']

#------------------------------------------------------------------------------
# Nearest-neighbor interpolation
#------------------------------------------------------------------------------


class NearestNDInterpolator(NDInterpolatorBase):
    """Nearest-neighbor interpolator in N > 1 dimensions.

    Methods
    -------
    __call__

    Parameters
    ----------
    x : (npoints, ndims) 2-D ndarray of floats
        Data point coordinates.
    y : (npoints, ...) N-D ndarray of float or complex
        Data values. The length of `y` along the first axis must be equal to
        the length of `x`.
    rescale : boolean, optional
        Rescale points to unit cube before performing interpolation.
        This is useful if some of the input dimensions have
        incommensurable units and differ by many orders of magnitude.

        .. versionadded:: 0.14.0
    tree_options : dict, optional
        Options passed to the underlying ``cKDTree``.

        .. versionadded:: 0.17.0

    See Also
    --------
    griddata :
        Interpolate unstructured D-D data.
    LinearNDInterpolator :
        Piecewise linear interpolator in N dimensions.
    CloughTocher2DInterpolator :
        Piecewise cubic, C1 smooth, curvature-minimizing interpolator in 2D.
    interpn : Interpolation on a regular grid or rectilinear grid.
    RegularGridInterpolator : Interpolator on a regular or rectilinear grid
                              in arbitrary dimensions (`interpn` wraps this
                              class).

    Notes
    -----
    Uses ``scipy.spatial.cKDTree``

    .. note:: For data on a regular grid use `interpn` instead.

    Examples
    --------
    We can interpolate values on a 2D plane:

    >>> from scipy.interpolate import NearestNDInterpolator
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> rng = np.random.default_rng()
    >>> x = rng.random(10) - 0.5
    >>> y = rng.random(10) - 0.5
    >>> z = np.hypot(x, y)
    >>> X = np.linspace(min(x), max(x))
    >>> Y = np.linspace(min(y), max(y))
    >>> X, Y = np.meshgrid(X, Y)  # 2D grid for interpolation
    >>> interp = NearestNDInterpolator(list(zip(x, y)), z)
    >>> Z = interp(X, Y)
    >>> plt.pcolormesh(X, Y, Z, shading='auto')
    >>> plt.plot(x, y, "ok", label="input point")
    >>> plt.legend()
    >>> plt.colorbar()
    >>> plt.axis("equal")
    >>> plt.show()

    """

    def __init__(self, x, y, rescale=False, tree_options=None):
        NDInterpolatorBase.__init__(self, x, y, rescale=rescale,
                                    need_contiguous=False,
                                    need_values=False)
        if tree_options is None:
            tree_options = dict()
        self.tree = cKDTree(self.points, **tree_options)
        self.values = np.asarray(y)

    def __call__(self, *args, **query_options):
        """
        Evaluate interpolator at given points.

        Parameters
        ----------
        x1, x2, ... xn : array-like of float
            Points where to interpolate data at.
            x1, x2, ... xn can be array-like of float with broadcastable shape.
            or x1 can be array-like of float with shape ``(..., ndim)``
        **query_options
            This allows ``eps``, ``p``, ``distance_upper_bound``, and ``workers``
            being passed to the cKDTree's query function to be explicitly set.
            See `scipy.spatial.cKDTree.query` for an overview of the different options.

            .. versionadded:: 1.12.0

        """
        # For the sake of enabling subclassing, NDInterpolatorBase._set_xi performs
        # some operations which are not required by NearestNDInterpolator.__call__, 
        # hence here we operate on xi directly, without calling a parent class function.
        xi = _ndim_coords_from_arrays(args, ndim=self.points.shape[1])
        xi = self._check_call_shape(xi)
        xi = self._scale_x(xi)

        # We need to handle two important cases:
        # (1) the case where xi has trailing dimensions (..., ndim), and
        # (2) the case where y has trailing dimensions
        # We will first flatten xi to deal with case (1),
        # do the computation in flattened array while retaining y's dimensionality,
        # and then reshape the interpolated values back to match xi's shape.

        # Flatten xi for the query
        xi_flat = xi.reshape(-1, xi.shape[-1])
        original_shape = xi.shape
        flattened_shape = xi_flat.shape

        # if distance_upper_bound is set to not be infinite,
        # then we need to consider the case where cKDtree
        # does not find any points within distance_upper_bound to return.
        # It marks those points as having infinte distance, which is what will be used
        # below to mask the array and return only the points that were deemed
        # to have a close enough neighbor to return something useful.
        dist, i = self.tree.query(xi_flat, **query_options)
        valid_mask = np.isfinite(dist)

        # create a holder interp_values array and fill with nans.
        if self.values.ndim > 1:
            interp_shape = flattened_shape[:-1] + self.values.shape[1:]
        else:
            interp_shape = flattened_shape[:-1]

        if np.issubdtype(self.values.dtype, np.complexfloating):
            interp_values = np.full(interp_shape, np.nan, dtype=self.values.dtype)
        else:
            interp_values = np.full(interp_shape, np.nan)

        interp_values[valid_mask] = self.values[i[valid_mask], ...]

        if self.values.ndim > 1:
            new_shape = original_shape[:-1] + self.values.shape[1:]
        else:
            new_shape = original_shape[:-1]
        interp_values = interp_values.reshape(new_shape)

        return interp_values


#------------------------------------------------------------------------------
# Convenience interface function
#------------------------------------------------------------------------------


def griddata(points, values, xi, method='linear', fill_value=np.nan,
             rescale=False):
    """
    Convenience function for interpolating unstructured data in multiple dimensions.

    Parameters
    ----------
    points : 2-D ndarray of floats with shape (n, D), or length D tuple of 1-D ndarrays with shape (n,).
        Data point coordinates.
    values : ndarray of float or complex, shape (n,)
        Data values.
    xi : 2-D ndarray of floats with shape (m, D), or length D tuple of ndarrays broadcastable to the same shape.
        Points at which to interpolate data.
    method : {'linear', 'nearest', 'cubic'}, optional
        Method of interpolation. One of

        ``nearest``
          return the value at the data point closest to
          the point of interpolation. See `NearestNDInterpolator` for
          more details.

        ``linear``
          tessellate the input point set to N-D
          simplices, and interpolate linearly on each simplex. See
          `LinearNDInterpolator` for more details.

        ``cubic`` (1-D)
          return the value determined from a cubic
          spline.

        ``cubic`` (2-D)
          return the value determined from a
          piecewise cubic, continuously differentiable (C1), and
          approximately curvature-minimizing polynomial surface. See
          `CloughTocher2DInterpolator` for more details.
    fill_value : float, optional
        Value used to fill in for requested points outside of the
        convex hull of the input points. If not provided, then the
        default is ``nan``. This option has no effect for the
        'nearest' method.
    rescale : bool, optional
        Rescale points to unit cube before performing interpolation.
        This is useful if some of the input dimensions have
        incommensurable units and differ by many orders of magnitude.

        .. versionadded:: 0.14.0

    Returns
    -------
    ndarray
        Array of interpolated values.

    See Also
    --------
    LinearNDInterpolator :
        Piecewise linear interpolator in N dimensions.
    NearestNDInterpolator :
        Nearest-neighbor interpolator in N dimensions.
    CloughTocher2DInterpolator :
        Piecewise cubic, C1 smooth, curvature-minimizing interpolator in 2D.
    interpn : Interpolation on a regular grid or rectilinear grid.
    RegularGridInterpolator : Interpolator on a regular or rectilinear grid
                              in arbitrary dimensions (`interpn` wraps this
                              class).

    Notes
    -----

    .. versionadded:: 0.9

    .. note:: For data on a regular grid use `interpn` instead.

    Examples
    --------

    Suppose we want to interpolate the 2-D function

    >>> import numpy as np
    >>> def func(x, y):
    ...     return x*(1-x)*np.cos(4*np.pi*x) * np.sin(4*np.pi*y**2)**2

    on a grid in [0, 1]x[0, 1]

    >>> grid_x, grid_y = np.mgrid[0:1:100j, 0:1:200j]

    but we only know its values at 1000 data points:

    >>> rng = np.random.default_rng()
    >>> points = rng.random((1000, 2))
    >>> values = func(points[:,0], points[:,1])

    This can be done with `griddata` -- below we try out all of the
    interpolation methods:

    >>> from scipy.interpolate import griddata
    >>> grid_z0 = griddata(points, values, (grid_x, grid_y), method='nearest')
    >>> grid_z1 = griddata(points, values, (grid_x, grid_y), method='linear')
    >>> grid_z2 = griddata(points, values, (grid_x, grid_y), method='cubic')

    One can see that the exact result is reproduced by all of the
    methods to some degree, but for this smooth function the piecewise
    cubic interpolant gives the best results:

    >>> import matplotlib.pyplot as plt
    >>> plt.subplot(221)
    >>> plt.imshow(func(grid_x, grid_y).T, extent=(0,1,0,1), origin='lower')
    >>> plt.plot(points[:,0], points[:,1], 'k.', ms=1)
    >>> plt.title('Original')
    >>> plt.subplot(222)
    >>> plt.imshow(grid_z0.T, extent=(0,1,0,1), origin='lower')
    >>> plt.title('Nearest')
    >>> plt.subplot(223)
    >>> plt.imshow(grid_z1.T, extent=(0,1,0,1), origin='lower')
    >>> plt.title('Linear')
    >>> plt.subplot(224)
    >>> plt.imshow(grid_z2.T, extent=(0,1,0,1), origin='lower')
    >>> plt.title('Cubic')
    >>> plt.gcf().set_size_inches(6, 6)
    >>> plt.show()

    """ # numpy/numpydoc#87  # noqa: E501

    points = _ndim_coords_from_arrays(points)

    if points.ndim < 2:
        ndim = points.ndim
    else:
        ndim = points.shape[-1]

    if ndim == 1 and method in ('nearest', 'linear', 'cubic'):
        from ._interpolate import interp1d
        points = points.ravel()
        if isinstance(xi, tuple):
            if len(xi) != 1:
                raise ValueError("invalid number of dimensions in xi")
            xi, = xi
        # Sort points/values together, necessary as input for interp1d
        idx = np.argsort(points)
        points = points[idx]
        values = values[idx]
        if method == 'nearest':
            fill_value = 'extrapolate'
        ip = interp1d(points, values, kind=method, axis=0, bounds_error=False,
                      fill_value=fill_value)
        return ip(xi)
    elif method == 'nearest':
        ip = NearestNDInterpolator(points, values, rescale=rescale)
        return ip(xi)
    elif method == 'linear':
        ip = LinearNDInterpolator(points, values, fill_value=fill_value,
                                  rescale=rescale)
        return ip(xi)
    elif method == 'cubic' and ndim == 2:
        ip = CloughTocher2DInterpolator(points, values, fill_value=fill_value,
                                        rescale=rescale)
        return ip(xi)
    else:
        raise ValueError(
            f"Unknown interpolation method {method!r} for {ndim} dimensional data"
        )



import pytest
import numpy as np

from scipy._lib._array_api import make_xp_test_case, xp_default_dtype, is_jax
from scipy._lib._array_api_no_0d import xp_assert_close
from scipy import stats
from scipy.stats._axis_nan_policy import SmallSampleWarning


@make_xp_test_case(stats.chatterjeexi)
class TestChatterjeeXi:
    @pytest.mark.parametrize('case', [
        dict(y_cont=True, statistic=-0.303030303030303, pvalue=0.9351329808526656),
        dict(y_cont=False, statistic=0.07407407407407396, pvalue=0.3709859367123997)])
    @pytest.mark.parametrize('dtype', ['float32', 'float64', None])
    def test_against_R_XICOR(self, case, dtype, xp):
        # Test against R package XICOR, e.g.
        # library(XICOR)
        # options(digits=16)
        # x = c(0.11027287231363914, 0.8154770102474279, 0.7073943466920335,
        #       0.6651317324378386, 0.6905752850115503, 0.06115250587536558,
        #       0.5209906494474178, 0.3155763519785274, 0.18405731803625924,
        #       0.8613557911541495)
        # y = c(0.8402081904493103, 0.5946972833914318, 0.23481606164114155,
        #       0.49754786197715384, 0.9146460831206026, 0.5848057749217579,
        #       0.7620801065573549, 0.31410063302647495, 0.7935620302236199,
        #       0.5423085761365468)
        # xicor(x, y, ties=FALSE, pvalue=TRUE)
        if dtype == 'float32' and np.__version__ < "2":
            pytest.skip("Scalar dtypes only respected after NEP 50.")
        dtype = xp_default_dtype(xp) if dtype is None else getattr(xp, dtype)
        rng = np.random.default_rng(25982435982346983)
        x = rng.random(size=10)
        y = (rng.random(size=10) if case['y_cont']
             else rng.integers(0, 5, size=10))

        x, y = xp.asarray(x, dtype=dtype), xp.asarray(y, dtype=dtype)
        res = stats.chatterjeexi(x, y, y_continuous=case['y_cont'])

        xp_assert_close(res.statistic, xp.asarray(case['statistic'], dtype=dtype))
        xp_assert_close(res.pvalue, xp.asarray(case['pvalue'], dtype=dtype))

    @pytest.mark.parametrize('y_continuous', (False, True))
    def test_permutation_asymptotic(self, y_continuous):
        # XICOR doesn't seem to perform the permutation test as advertised, so
        # compare the result of a permutation test against an asymptotic test.
        rng = np.random.default_rng(2524579827426)
        n = np.floor(rng.uniform(100, 150)).astype(int)
        shape = (2, n)
        x = rng.random(size=shape)
        y = (rng.random(size=shape) if y_continuous
             else rng.integers(0, 10, size=shape))
        method = stats.PermutationMethod(rng=rng)
        res = stats.chatterjeexi(x, y, method=method,
                                 y_continuous=y_continuous, axis=-1)
        ref = stats.chatterjeexi(x, y, y_continuous=y_continuous, axis=-1)
        np.testing.assert_allclose(res.statistic, ref.statistic, rtol=1e-15)
        np.testing.assert_allclose(res.pvalue, ref.pvalue, rtol=2e-2)

    def test_input_validation(self, xp):
        rng = np.random.default_rng(25932435798274926)
        x, y = rng.random(size=(2, 10))
        x, y = xp.asarray(x), xp.asarray(y)

        message = 'Array shapes are incompatible for broadcasting.|Incompatible shapes'
        with pytest.raises((ValueError, TypeError), match=message):
            stats.chatterjeexi(x, y[:-1])

        if not is_jax(xp):
            # jax misses out on some input validation from _axis_nan_policy decorator
            message = '...axis 10 is out of bounds for array...|out of range'
            with pytest.raises((ValueError, IndexError), match=message):
                stats.chatterjeexi(x, y, axis=10)

        message = '`y_continuous` must be boolean.'
        with pytest.raises(ValueError, match=message):
            stats.chatterjeexi(x, y, y_continuous='a herring')

        message = "`method` must be 'asymptotic' or"
        with pytest.raises(ValueError, match=message):
            stats.chatterjeexi(x, y, method='ekki ekii')

    @pytest.mark.skip_xp_backends('jax.numpy', reason='no SmallSampleWarning (lazy)')
    def test_special_cases(self, xp):
        message = 'One or more sample arguments is too small...'
        with pytest.warns(SmallSampleWarning, match=message):
            res = stats.chatterjeexi(xp.asarray([1]), xp.asarray([2]))

        assert xp.isnan(res.statistic)
        assert xp.isnan(res.pvalue)


@make_xp_test_case(stats.spearmanrho)
class TestSpearmanRho:
    @pytest.mark.parametrize('alternative, statistic, pvalue', [
        ('two-sided', -0.2727272727272727, 0.4458383415428),
        ('greater', -0.2727272727272727, 0.7770808292286),
        ('less', -0.2727272727272727, 0.2229191707714),
    ])
    @pytest.mark.parametrize('dtype', ['float32', 'float64', None])
    def test_against_R_cor_test(self, alternative, statistic, pvalue, dtype, xp):
        # Test against R cor.test, e.g.
        # options(digits=16)
        # x = c(0.11027287231363914, 0.8154770102474279, 0.7073943466920335,
        #       0.6651317324378386, 0.6905752850115503, 0.06115250587536558,
        #       0.5209906494474178, 0.3155763519785274, 0.18405731803625924,
        #       0.8613557911541495)
        # y = c(0.8402081904493103, 0.5946972833914318, 0.23481606164114155,
        #       0.49754786197715384, 0.9146460831206026, 0.5848057749217579,
        #       0.7620801065573549, 0.31410063302647495, 0.7935620302236199,
        #       0.5423085761365468)
        # cor.test(x, y, method='spearman', alternative='t', exact=FALSE)
        if dtype == 'float32' and np.__version__ < "2":
            pytest.skip("Scalar dtypes only respected after NEP 50.")
        dtype = xp_default_dtype(xp) if dtype is None else getattr(xp, dtype)
        rng = np.random.default_rng(25982435982346983)
        x = rng.random(size=10)
        y = rng.random(size=10)

        x, y = xp.asarray(x, dtype=dtype), xp.asarray(y, dtype=dtype)
        res = stats.spearmanrho(x, y, alternative=alternative)

        xp_assert_close(res.statistic, xp.asarray(statistic, dtype=dtype))
        xp_assert_close(res.pvalue, xp.asarray(pvalue, dtype=dtype))


    @pytest.mark.parametrize('alternative, statistic, pvalue', [
        ('two-sided', -0.4857142857142857, 0.3555555555556),
        ('greater', -0.4857142857142857, 0.8513888888889),
        ('less', -0.4857142857142857, 0.1777777777778),
    ])
    def test_against_R_cor_test_exact(self, alternative, statistic, pvalue):
        xp = np  # will test for multiple backends when gh-23772 merges
        # Test against R cor.test exact=TRUE, e.g.
        # options(digits=16)
        # x = c(0.11027287231363914, 0.8154770102474279, 0.7073943466920335,
        #       0.6651317324378386, 0.6905752850115503, 0.06115250587536558)
        # y = c(0.5209906494474178, 0.3155763519785274, 0.18405731803625924,
        #       0.8613557911541495, 0.8402081904493103, 0.5946972833914318)
        # cor.test(x, y, method='spearman', alternative='t', exact=TRUE)
        rng = np.random.default_rng(25982435982346983)
        x = rng.random(size=6)
        y = rng.random(size=6)
        x, y = xp.asarray(x), xp.asarray(y)

        method = stats.PermutationMethod()
        res = stats.spearmanrho(x, y, method=method, alternative=alternative)

        xp_assert_close(res.statistic, xp.asarray(statistic))
        xp_assert_close(res.pvalue, xp.asarray(pvalue))


    @pytest.mark.parametrize('alternative', ('two-sided', 'greater', 'less'))
    @pytest.mark.parametrize('n', [9, 99, 999])
    def test_against_scipy_spearmanr(self, alternative, n, xp):
        rng = np.random.default_rng(5982435982346983)
        x = rng.integers(n//2, size=n)
        y = rng.integers(n//2, size=n)
        dtype = xp_default_dtype(xp)

        ref = stats.spearmanr(x, y, alternative=alternative)
        ref_statistic = xp.asarray(ref.statistic, dtype=dtype)
        ref_pvalue = xp.asarray(ref.pvalue, dtype=dtype)

        x = xp.asarray(x, dtype=dtype)
        y = xp.asarray(y, dtype=dtype)
        res = stats.spearmanrho(x, y, alternative=alternative)

        xp_assert_close(res.statistic, ref_statistic)
        xp_assert_close(res.pvalue, ref_pvalue)


    @pytest.mark.parametrize('axis', [-1, 0, 1])
    def test_other_backends_nd(self, axis, xp):
        # NumPy n-d behavior is tested by `test_axis_nan_policy`;
        # check other backends against it.
        rng = np.random.default_rng(8243598346983259)
        shape = (8, 9, 10)
        x = rng.standard_normal(size=shape)
        y = rng.standard_normal(size=shape)
        res = stats.spearmanrho(xp.asarray(x), xp.asarray(y), axis=axis)
        ref = stats.spearmanrho(x, y, axis=axis)
        xp_assert_close(res.statistic, xp.asarray(ref.statistic), atol=1e-16)
        xp_assert_close(res.pvalue, xp.asarray(ref.pvalue), atol=1e-16)


    def test_input_validation(self, xp):
        rng = np.random.default_rng(25932435798274926)
        x, y = rng.random(size=(2, 10))
        x, y = xp.asarray(x), xp.asarray(y)

        msg = 'incompatible for broadcasting|Incompatible shapes|must be broadcastable'
        with pytest.raises((ValueError, TypeError), match=msg):
            stats.spearmanrho(x, y[:-1])

        if not is_jax(xp):
            # jax misses out on some input validation from _axis_nan_policy decorator
            message = '...axis 10 is out of bounds for array...|out of range'
            with pytest.raises((ValueError, IndexError), match=message):
                stats.spearmanrho(x, y, axis=10)

        message = "`alternative` must be 'less', 'greater', or 'two-sided'."
        with pytest.raises(ValueError, match=message):
            stats.spearmanrho(x, y, alternative='alternative')

        message = ("`method` must be...")
        with pytest.raises(ValueError, match=message):
            stats.spearmanrho(x, y, method='method')


    @pytest.mark.skip_xp_backends('jax.numpy', reason='no SmallSampleWarning (lazy)')
    def test_special_cases(self, xp):
        def check_nan(res):
            assert xp.isnan(res.statistic)
            assert xp.isnan(res.pvalue)

        message = 'One or more sample arguments is too small...'
        with pytest.warns(SmallSampleWarning, match=message):
            res = stats.spearmanrho(xp.asarray([1]), xp.asarray([2]))
            check_nan(res)

        x = xp.asarray([1, 1, 1, 1, 1])
        y = xp.asarray([1, 2, 3, 4, 5])
        message = 'An input array is constant; the correlation coefficient...'
        with pytest.warns(stats.ConstantInputWarning, match=message):
            res = stats.spearmanrho(x, y)
            check_nan(res)
        with pytest.warns(stats.ConstantInputWarning, match=message):
            res = stats.spearmanrho(y, x)
            check_nan(res)



"""Test exported names"""

import builtins

import numpy as np
import pytest

from array_api_compat._internal import clone_module

from ._helpers import wrapped_libraries

NAMES = {
    "": [
        # Inspection
        "__array_api_version__",
        "__array_namespace_info__",
        # Submodules
        "fft",
        "linalg",
        # Constants
        "e",
        "inf",
        "nan",
        "newaxis",
        "pi",
        # Creation Functions
        "arange",
        "asarray",
        "empty",
        "empty_like",
        "eye",
        "from_dlpack",
        "full",
        "full_like",
        "linspace",
        "meshgrid",
        "ones",
        "ones_like",
        "tril",
        "triu",
        "zeros",
        "zeros_like",
        # Data Type Functions
        "astype",
        "can_cast",
        "finfo",
        "iinfo",
        "isdtype",
        "result_type",
        # Data Types
        "bool",
        "int8",
        "int16",
        "int32",
        "int64",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "float32",
        "float64",
        "complex64",
        "complex128",
        # Elementwise Functions
        "abs",
        "acos",
        "acosh",
        "add",
        "asin",
        "asinh",
        "atan",
        "atan2",
        "atanh",
        "bitwise_and",
        "bitwise_left_shift",
        "bitwise_invert",
        "bitwise_or",
        "bitwise_right_shift",
        "bitwise_xor",
        "ceil",
        "clip",
        "conj",
        "copysign",
        "cos",
        "cosh",
        "divide",
        "equal",
        "exp",
        "expm1",
        "floor",
        "floor_divide",
        "greater",
        "greater_equal",
        "hypot",
        "imag",
        "isfinite",
        "isinf",
        "isnan",
        "less",
        "less_equal",
        "log",
        "log1p",
        "log2",
        "log10",
        "logaddexp",
        "logical_and",
        "logical_not",
        "logical_or",
        "logical_xor",
        "maximum",
        "minimum",
        "multiply",
        "negative",
        "nextafter",
        "not_equal",
        "positive",
        "pow",
        "real",
        "reciprocal",
        "remainder",
        "round",
        "sign",
        "signbit",
        "sin",
        "sinh",
        "square",
        "sqrt",
        "subtract",
        "tan",
        "tanh",
        "trunc",
        # Indexing Functions
        "take",
        "take_along_axis",
        # Linear Algebra Functions
        "matmul",
        "matrix_transpose",
        "tensordot",
        "vecdot",
        # Manipulation Functions
        "broadcast_arrays",
        "broadcast_to",
        "concat",
        "expand_dims",
        "flip",
        "moveaxis",
        "permute_dims",
        "repeat",
        "reshape",
        "roll",
        "squeeze",
        "stack",
        "tile",
        "unstack",
        # Searching Functions
        "argmax",
        "argmin",
        "count_nonzero",
        "nonzero",
        "searchsorted",
        "where",
        # Set Functions
        "unique_all",
        "unique_counts",
        "unique_inverse",
        "unique_values",
        # Sorting Functions
        "argsort",
        "sort",
        # Statistical Functions
        "cumulative_prod",
        "cumulative_sum",
        "max",
        "mean",
        "min",
        "prod",
        "std",
        "sum",
        "var",
        # Utility Functions
        "all",
        "any",
        "diff",
    ],
    "fft": [
        "fft",
        "ifft",
        "fftn",
        "ifftn",
        "rfft",
        "irfft",
        "rfftn",
        "irfftn",
        "hfft",
        "ihfft",
        "fftfreq",
        "rfftfreq",
        "fftshift",
        "ifftshift",
    ],
    "linalg": [
        "cholesky",
        "cross",
        "det",
        "diagonal",
        "eigh",
        "eigvalsh",
        "inv",
        "matmul",
        "matrix_norm",
        "matrix_power",
        "matrix_rank",
        "matrix_transpose",
        "outer",
        "pinv",
        "qr",
        "slogdet",
        "solve",
        "svd",
        "svdvals",
        "tensordot",
        "trace",
        "vecdot",
        "vector_norm",
    ],
}

XFAILS = {
    ("numpy", ""): ["from_dlpack"] if np.__version__ < "1.23" else [],
    ("dask.array", ""): ["from_dlpack", "take_along_axis"],
    ("dask.array", "linalg"): [
        "cross",
        "det",
        "eigh",
        "eigvalsh",
        "matrix_power",
        "pinv",
        "slogdet",
    ],
}


def all_names(mod):
    """Return all names available in a module."""
    objs = {}
    clone_module(mod.__name__, objs)
    return set(objs)


def get_mod(library, module, *, compat):
    if compat:
        library = f"array_api_compat.{library}"
    xp = pytest.importorskip(library)
    return getattr(xp, module) if module else xp


@pytest.mark.parametrize("module", list(NAMES))
@pytest.mark.parametrize("library", wrapped_libraries)
def test_array_api_names(library, module):
    """Test that __all__ isn't missing any exports
    dictated by the Standard.
    """
    mod = get_mod(library, module, compat=True)
    missing = set(NAMES[module]) - all_names(mod)
    xfail = set(XFAILS.get((library, module), []))
    xpass = xfail - missing
    fails = missing - xfail
    assert not xpass, f"Names in XFAILS are defined: {xpass}"
    assert not fails, f"Missing exports: {fails}"


@pytest.mark.parametrize("module", list(NAMES))
@pytest.mark.parametrize("library", wrapped_libraries)
def test_compat_doesnt_hide_names(library, module):
    """The base namespace can have more names than the ones explicitly exported
    by array-api-compat. Test that we're not suppressing them.
    """
    bare_mod = get_mod(library, module, compat=False)
    compat_mod = get_mod(library, module, compat=True)

    missing = all_names(bare_mod) - all_names(compat_mod)
    missing = {name for name in missing if not name.startswith("_")}
    assert not missing, f"Non-Array API names have been hidden: {missing}"


@pytest.mark.parametrize("module", list(NAMES))
@pytest.mark.parametrize("library", wrapped_libraries)
def test_compat_doesnt_add_names(library, module):
    """Test that array-api-compat isn't adding names to the namespace
    besides those defined by the Array API Standard.
    """
    bare_mod = get_mod(library, module, compat=False)
    compat_mod = get_mod(library, module, compat=True)

    aapi_names = set(NAMES[module])
    spurious = all_names(compat_mod) - all_names(bare_mod) - aapi_names
    # Quietly ignore *Result dataclasses
    spurious = {name for name in spurious if not name.endswith("Result")}
    assert not spurious, (
        f"array-api-compat is adding non-Array API names: {spurious}"
    )


@pytest.mark.parametrize(
    "name", [name for name in NAMES[""] if hasattr(builtins, name)]
)
@pytest.mark.parametrize("library", wrapped_libraries)
def test_builtins_collision(library, name):
    """Test that xp.bool is not accidentally builtins.bool, etc."""
    xp = pytest.importorskip(f"array_api_compat.{library}")
    assert getattr(xp, name) is not getattr(builtins, name)



"""Test xsf functions work on GPU through CuPy.

Beyond cupy and pytest, also requires xsref package to be installed.
Should be run with pytest-isolate or pytest-forked to isolate tests
in separate processes since memory corruption on GPU can cause
failures to occur in unrelated tests.

TODO:
Set this up to run through pixi, and to run in CI.
"""

import os
import numpy as np
import polars as pl
import pytest
import shutil
import tempfile

from glob import glob
from packaging.version import Version
from pathlib import Path

from xsref.float_tools import extended_relative_error
from xsref.tables import get_input_rows, get_output_rows, get_in_out_types


#------------------------------------------------------------------------------
# Check if a module is present to be used in tests
#
# Copied from
# https://github.com/scipy/scipy/blob/1cbfa1c894557041d9825d6754a2c48fc3bec484/scipy/special/_testutils.py
#------------------------------------------------------------------------------

class MissingModule:
    def __init__(self, name):
        self.name = name


def check_version(module, min_ver):
    if type(module) is MissingModule:
        return pytest.mark.skip(reason=f"{module.name} is not installed")
    return pytest.mark.skipif(
        Version(module.__version__) < Version(min_ver),
        reason=f"{module.__name__} version >= {min_ver} required"
    )

#------------------------------------------------------------------------------

try:
    import cupy  # type: ignore
except (ImportError, AttributeError):
    cupy = MissingModule('cupy')


@pytest.fixture(scope="function", autouse=True)
def manage_cupy_cache():
    # Temporarily change cupy kernel cache location so kernel cache will not be polluted
    # by these tests. Remove temporary cache in teardown.
    temp_cache_dir = tempfile.mkdtemp()
    original_cache_dir = os.environ.get('CUPY_CACHE_DIR', None)
    os.environ['CUPY_CACHE_DIR'] = temp_cache_dir

    yield

    if original_cache_dir is not None:
        os.environ['CUPY_CACHE_DIR'] = original_cache_dir
    else:
        del os.environ['CUPY_CACHE_DIR']
    shutil.rmtree(temp_cache_dir)


def _get_cols_helper(table_path, xp):
    # This is more complicated than need be due to an oversight in xsref
    # functions for getting tables. The table metadata for input, output,
    # and err tables has the input and output types, but does not have
    # the column types of the current table, and does not contain the info
    # of whether current table is an input, output, or err table. It also
    # provides separate functions to get input rows from an input table,
    # output rows from an output table, and provides no function to get
    # rows from an err table (though this last task is easy to do because
    # there are no complex types involved in an err table).
    # TODO: Update xsref tables to contain the table's column types
    # in the metadata, and provide a single function for getting rows
    # from tables as lists of tuples that uses this metadata.
    table_name = table_path.name.lower()
    if table_name.startswith("in_"):
        rows = get_input_rows(table_path)
    elif table_name.startswith("out_"):
        rows = get_output_rows(table_path)
    elif table_name.startswith("err_"):
        table = pl.read_parquet(table_path)
        rows = table.to_numpy()
    result = tuple(xp.asarray(col) for col in zip(*rows))
    if len(result) == 1:
        result = result[0]
    return result


def get_cols_as_cupy(table_path):
    return _get_cols_helper(table_path, cupy)


def get_cols_as_numpy(table_path):
    return _get_cols_helper(table_path, np)


HERE = Path(__file__)


def get_tables_for_func(func_name):
    tables_path = (
        HERE.parent.parent.resolve() / "xsref" / "tables" / "scipy_special_tests"
    )
    tables_path /= func_name
    input_tables = list(tables_path.glob("In_*.parquet"))
    output_tables = [
        path.parent / path.name.replace("In_", "Out_") for path in input_tables
    ]
    err_tables = []
    for path in input_tables:
        types = path.name.removesuffix(".parquet").replace("In_", "")
        name = f"Err_{types}_other.parquet"
        err_tables.append(path.parent / name)
    return list(zip(input_tables, output_tables, err_tables))


def get_preamble(header):
    header_path = (HERE.parent.parent / "include" / Path(header)).resolve()
    return f'#include "{header_path}"'


@pytest.mark.usefixtures("manage_cupy_cache")
@check_version(cupy, "13.0.0")
class TestCuPy:
    def _adjust_tol(self, tol, *, wiggle=16):
        return wiggle * np.maximum(tol, np.finfo(tol.dtype).eps)

    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("beta")
    )
    def test_beta(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        beta = cupy._core.create_ufunc(
            "cupyx_scipy_beta",
            ("ll->d", "LL->d", "ee->d", "ff->f", "dd->d"),
            "out0 = out0_type(xsf::beta(in0, in1));",
            preamble=get_preamble("xsf/beta.h"),
        )

        a, b = get_cols_as_cupy(input_path)
        out = cupy.asnumpy(beta(a, b))

        desired = get_cols_as_numpy(output_path)
        rtol = get_cols_as_numpy(tol_path)
        assert np.all(
            extended_relative_error(out, desired) <= self._adjust_tol(rtol)
        )

    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("binom")
    )
    def test_binom(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        binom = cupy._core.create_ufunc(
            "cupyx_scipy_binom",
            ("ff->f", "dd->d"),
            "out0 = out0_type(xsf::binom(in0, in1));",
            preamble=get_preamble("xsf/binom.h"),
        )

        n, k = get_cols_as_cupy(input_path)
        out = cupy.asnumpy(binom(n, k))

        desired = get_cols_as_numpy(output_path)
        rtol = get_cols_as_numpy(tol_path)
        assert np.all(
            extended_relative_error(out, desired) <= self._adjust_tol(rtol)
        )

    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("digamma")
    )
    def test_digamma(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        digamma = cupy._core.create_ufunc(
            'cupyx_scipy_special_digamma',
            (
                ('l->d', 'out0 = xsf::digamma(double(in0))'),
                ('e->d', 'out0 = xsf::digamma(double(in0))'),
                'f->f',
                'd->d',
                'F->F',
                'D->D',
            ),
            'out0 = xsf::digamma(in0)',
            preamble=get_preamble("xsf/digamma.h")
        )

        x = get_cols_as_cupy(input_path)
        out = cupy.asnumpy(digamma(x))

        desired = get_cols_as_numpy(output_path)
        rtol = get_cols_as_numpy(tol_path)
        assert np.all(
            extended_relative_error(out, desired) <= self._adjust_tol(rtol)
        )


    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("ellipkinc")
    )
    def test_ellipkinc(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        ellipkinc = cupy._core.create_ufunc(
            'cupyx_scipy_special_ellipkinc', ('ff->f', 'dd->d'),
            'out0 = xsf::cephes::ellik(in0, in1)',
            preamble=get_preamble("xsf/cephes/ellik.h"),
        )

        phi, m = get_cols_as_cupy(input_path)
        out = cupy.asnumpy(ellipkinc(phi, m))

        desired = get_cols_as_numpy(output_path)
        rtol = get_cols_as_numpy(tol_path)
        assert np.all(
            extended_relative_error(out, desired) <= self._adjust_tol(rtol)
        )

    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("ellipeinc")
    )
    def test_ellipeinc(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        ellipeinc = cupy._core.create_ufunc(
            'cupyx_scipy_special_ellipeinc', ('ff->f', 'dd->d'),
            'out0 = xsf::cephes::ellie(in0, in1)',
            preamble=get_preamble("xsf/cephes/ellie.h"),
        )

        phi, m = get_cols_as_cupy(input_path)
        out = cupy.asnumpy(ellipeinc(phi, m))

        desired = get_cols_as_numpy(output_path)
        rtol = get_cols_as_numpy(tol_path)
        assert np.all(
            extended_relative_error(out, desired) <= self._adjust_tol(rtol)
        )

    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("expn")
    )
    def test_expn(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        expn = cupy._core.create_ufunc(
            'cupyx_scipy_special_expn',
            ('ff->f', 'dd->d'),
            'out0 = xsf::cephes::expn(in0, in1)',
            preamble=get_preamble("xsf/cephes/expn.h"),
        )

        x, n = get_cols_as_cupy(input_path)
        out = cupy.asnumpy(expn(x, n))

        desired = get_cols_as_numpy(output_path)
        rtol = get_cols_as_numpy(tol_path)
        assert np.all(
            extended_relative_error(out, desired) <= self._adjust_tol(rtol)
        )

    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("gdtrib")
    )
    @pytest.mark.xfail(reason="Requires cpp_std >= 17")
    def test_gdtrib(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        gdtrib = cupy._core.create_ufunc(
            'cupyx_scipy_special_gdtrib',
            ('fff->f', 'ddd->d'),
            'out0 = xsf::gdtrib(in0, in1, in2)',
            preamble=get_preamble("xsf/cdflib.h"),
        )

        a, p, x = get_cols_as_cupy(input_path)
        out = cupy.asnumpy(gdtrib(a, p, x))

        desired = get_cols_as_numpy(output_path)
        rtol = get_cols_as_numpy(tol_path)
        assert np.all(
            extended_relative_error(out, desired) <= self._adjust_tol(rtol)
        )

    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("hyp2f1")
    )
    def test_hyp2f1(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        hyp2f1 = cupy._core.create_ufunc(
            'cupyx_scipy_special_hyp2f1',
            ('ffff->f', 'dddd->d', 'fffF->F', 'dddD->D'),
            'out0 = xsf::hyp2f1(in0, in1, in2, in3)',
            preamble=get_preamble("xsf/hyp2f1.h"),
        )

        a, b, c, z = get_cols_as_cupy(input_path)
        if not cupy.iscomplexobj(z):
            pytest.xfail(
                "Real valued hyp2f1 currently broken on GPU due to use of"
                " recursion."
            )
        out = cupy.asnumpy(hyp2f1(a, b, c, z))

        desired = get_cols_as_numpy(output_path)
        rtol = get_cols_as_numpy(tol_path)
        assert np.all(
            extended_relative_error(out, desired) <= self._adjust_tol(rtol)
        )

    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("lambertw")
    )
    def test_lambertw(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        _lambertw_scalar = cupy._core.create_ufunc(
            "cupyx_scipy_lambertw_scalar",
            ("Dld->D", "Fif->f"),
            "out0 = xsf::lambertw(in0, in1, in2)",
            preamble=get_preamble("xsf/lambertw.h"),
        )

        # A parameter called tol, not to be confused with the rtol for assessing
        # accuracy.
        z, k, tol = get_cols_as_cupy(input_path)
        out = cupy.asnumpy(_lambertw_scalar(z, k, tol))

        desired = get_cols_as_numpy(output_path)
        rtol = get_cols_as_numpy(tol_path)
        assert np.all(
            extended_relative_error(out, desired) <= self._adjust_tol(rtol)
        )

    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("sici")
    )
    def test_sici(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        sici = cupy._core.create_ufunc(
            'cupyx_scipy_special_sici',
            (
                (
                    'f->ff',
                    '''
                    float si, ci;
                    xsf::sici(in0, si, ci);
                    out0 = si; out1 = ci;
                    ''',
                ),
            (
                'd->dd',
                '''
                double si, ci;
                xsf::sici(in0, si, ci);
                out0 = si; out1 = ci;
                ''',
            ),
                (
                    'F->FF',
                    '''
                    complex<float> si, ci;
                    xsf::sici(in0, si, ci);
                out0 = si; out1 = ci;
                    ''',
                ),
                (
                    'D->DD',
                    '''
                    complex<double> si, ci;
                    xsf::sici(in0, si, ci);
                    out0 = si; out1 = ci;
                    ''',
                ),
            ),
            preamble=get_preamble("xsf/sici.h"),
        )
            
        x = get_cols_as_cupy(input_path)
        if cupy.iscomplexobj(x):
            pytest.xfail("Known bug, returning nan instead of a complex infinity.")
        out0, out1 = map(cupy.asnumpy, sici(x))

        desired0, desired1 = get_cols_as_numpy(output_path)
        rtol0, rtol1 = get_cols_as_numpy(tol_path)
        error = extended_relative_error(out0, desired0)
        tol = self._adjust_tol(rtol0)
        assert np.all(
            extended_relative_error(out0, desired0) <= self._adjust_tol(rtol0)
        )
        assert np.all(
            extended_relative_error(out1, desired1) <= self._adjust_tol(rtol1)
        )

    @pytest.mark.parametrize(
        "tables_paths", get_tables_for_func("shichi")
    )
    def test_shichi(self, tables_paths):
        input_path, output_path, tol_path = tables_paths
        shichi = cupy._core.create_ufunc(
            'cupyx_scipy_special_shichi',
            (
                (
                    'f->ff',
                    '''
                    float shi, chi;
                    xsf::shichi(in0, shi, chi);
                    out0 = shi; out1 = chi;
                    ''',
                ),
                (
                    'd->dd',
                    '''
                    double shi, chi;
                    xsf::shichi(in0, shi, chi);
                    out0 = shi; out1 = chi;
                    ''',
                ),
                (
                    'F->FF',
                    '''
                    complex<float> shi, chi;
                    xsf::shichi(in0, shi, chi);
                    out0 = shi; out1 = chi;
                    ''',
                ),
                (
                    'D->DD',
                    '''
                    complex<double> shi, chi;
                    xsf::shichi(in0, shi, chi);
                    out0 = shi; out1 = chi;
                    ''',
                ),
            ),
            preamble=get_preamble("xsf/sici.h"),
        )
            
        x = get_cols_as_cupy(input_path)
        if cupy.iscomplexobj(x):
            pytest.xfail("Known bug, returning nan instead of a complex infinity.")
        out0, out1 = map(cupy.asnumpy, shichi(x))

        desired0, desired1 = get_cols_as_numpy(output_path)
        rtol0, rtol1 = get_cols_as_numpy(tol_path)
        assert np.all(
            extended_relative_error(out0, desired0) <= self._adjust_tol(rtol0)
        )
        assert np.all(
            extended_relative_error(out1, desired1) <= self._adjust_tol(rtol1)
        )
